// Gaia Sky port of shadertoy shader "[Fork] Helix Nebula" by izutionix with
// colors changed to resemble the Cat's Eye Nebula.
// Licensed under CC-BY-NC-SA.
// Source: https://www.shadertoy.com/view/cdK3Wy
#version 330 core

// Uniforms which are always available
uniform mat4 u_worldTrans;
uniform vec3 u_cameraPos;
uniform float u_bodySize;
uniform vec2 u_cameraNearFar;
uniform float u_cameraK;
uniform float u_time;
uniform float u_simuTime;
// Unit conversion.
uniform float u_kmToU;

#ifdef ssrFlag
#include <shader/lib/ssr.frag.glsl>
#endif // ssrFlag

// INPUT
struct VertexData {
    vec2 texCoords;
    vec3 normal;
    vec3 viewDir;
    vec3 ambientLight;
    float opacity;
    vec4 color;
    vec3 fragPosWorld;
    mat3 tbn;
};
in VertexData v_data;

// OUTPUT
layout(location = 0) out vec4 fragColor;
layout(location = 1) out vec4 layerBuffer;

#include <shader/lib/logdepthbuff.glsl>
#include <$data/catalog-nebulae/shader/lib/nebulae-lib.glsl>

#define DIST_SCALE (u_kmToU * 1.0e6)
#define FACTOR (1.0 / DIST_SCALE * 2.0e-7)

#define ITERATIONS 60.0

float Disk(vec3 p, vec3 t) {
    vec2 q = vec2(length2(p.xy) - t.x, p.z * 0.5);
    return max(length8(q) - t.y, abs(p.z) - t.z);
}

//==============================================================
// otaviogood's noise from https://www.shadertoy.com/view/ld2SzK
//--------------------------------------------------------------
// This spiral noise works by successively adding and rotating sin waves while increasing frequency.
// It should work the same on all computers since it's not based on a hash function like some other noises.
// It can be much faster than other noise functions if you're ok with some repetition.
const float nudge = 0.6; // Size of perpendicular vector
float normalizer = 1.0 / sqrt(1.2 + nudge * nudge); // Pythagorean theorem on that perpendicular to maintain scale
float SpiralNoiseC(vec3 p) {
    float n = 0.0; // Noise amount
    float iter = 1.9;
    for (int i = 0; i < 6; i++) {
        // Add sin and cos scaled inverse with the frequency
        n += -abs(sin(p.y * iter) + cos(p.x * iter)) / iter; // Abs for a ridged look
        // Rotate by adding perpendicular and scaling down
        p.xy += vec2(p.y, -p.x) * nudge;
        p.xy *= normalizer;
        // Rotate on other axis
        p.xz += vec2(p.z, -p.x) * nudge;
        p.xz *= normalizer;
        // Increase the frequency
        iter *= 1.8033733;
    }
    return n;
}

float NebulaNoise(vec3 p) {
    float final = Disk(p.xzy, vec3(1.0, 2.2, 1.05));
    final += fbm(p * 20.0);
    final += SpiralNoiseC(p);

    return final;
}

float map(vec3 p) {
    // The Cat's Eye nebula is oriented sideways.
    vec4 q = get_quat_rotation(vec3(0.2, 0.1, 1.0), PI / 2.0);
    p = rotate(p, q);
    float NebNoise = abs(NebulaNoise(p / 0.645) * 0.35);
    return NebNoise + 0.098;
}

vec4 raymarch(vec3 camPos, vec3 rayDir, vec3 objPos, vec2 dists, out float s) {
    // Ray direction.
    vec3 rd = rayDir;
    // Ray origin.
    vec3 ro = camPos;

    // ld, td: local, total density
    // w: weighting factor
    float ld = 0.0, td = 0.0, w = 0.0;

    // t: length of the ray
    // d: distance function
    float d = 1.0;

    const float h = 0.1;

    vec4 sum = vec4(0.0);

    float min_dist = dists.x, max_dist = dists.y;

    if (min_dist >= 0.0 && max_dist >= 0.0) {
        float t = min_dist;

        // Raymarch loop
        for (int i = 0; i < ITERATIONS; i++) {
            vec3 pos = ro + t * rd;

            // Loop break conditions
            if (td > 0.9 || d < 0.02 * t || t > 15.0 || sum.a > 0.99) break;

            // Evaluate distance function
            float d = map(pos);

            // Change this string to control density
            d = max(d, 0.0);

            // Point light calculations
            vec3 ldst = vec3(0.0) - pos;
            float lDist = max(length(ldst), 0.001);

            // The color of light
            float _T = lDist * 2.3 + 2.6; // <-v endless tweaking
            vec3 lightColor = 0.2 + 0.6 * cos(_T * 0.84 + PI * 0.8 * vec3(0.4, 0.5, 0.5));

            // Star
            sum.rgb += (vec3(0.87, 0.5, 0.5) / (lDist * lDist * 10.0) / 40.0);
            // Bloom
            sum.rgb += (lightColor / exp(lDist * lDist * lDist * 0.08) / 20.0);

            if (d < h) {
                // Compute local density
                ld = h - d;

                // Compute weighting factor
                w = (1.0 - td) * ld;

                // Accumulate density
                td += w + 1.0 / 200.0;

                vec4 col = vec4(0.3, 0.3, 0.3, td);

                // Emission
                sum += sum.a * vec4(sum.rgb, 0.0) * 0.2;

                // Uniform scale density
                col.a *= 0.2;
                // Colour by alpha
                col.rgb *= col.a;
                // Alpha blend in contribution
                sum = sum + col * (1.0 - sum.a);

                if (s == 0.0) {
                    s = t;
                }
            }
            td += 1.0 / 90.0;
            // Optimize step size near the camera
            t += max(d * 0.1 * max(min(length(ldst), length(ro)), 1.0), 0.01);
        }
        td *= smoothstep(0.0, 0.8, td) * 0.8;

        // Simple scattering
        sum *= 1.0 / exp(ld * 0.2) * 0.3;
        sum = clamp(sum, 0.0, 1.0);
        sum.xyz = sum.xyz * sum.xyz * (3.0 - 2.0 * sum.xyz);
    }
    return vec4(sum.xyz, td);
}

void main() {
    // Directions and positions.
    // Object is alwyas at origin in our reference frame.
    vec3 camPos = u_cameraPos * FACTOR;
    vec3 rayDir = normalize(v_data.fragPosWorld);
    vec3 objPos = vec3(0.0, 0.0, 0.0);
    float radius = u_bodySize;

    vec2 dists = sphereIntersectionDistances(camPos, rayDir, objPos, radius);

    // Ray marching
    float d = 0.0;
    vec4 rmcol = raymarch(camPos, rayDir, objPos, dists, d);
    fragColor = rmcol;

    // Render outline
    // vec2 uv = v_data.texCoords;
    // if (uv.x < 0.005 || uv.x > 0.995 || uv.y < 0.005 || uv.y > 0.995) {
    //     d = length(v_data.fragPosWorld);
    //     fragColor = vec4(1.0, 0.0, 0.0, 1.0);
    // }

    // Logarithmic depth buffer.
    if (d > 0.0) {
        gl_FragDepth = getDepthValue(d, u_cameraNearFar.y, u_cameraK);
    }

    #ifdef ssrFlag
    ssrBuffers();
    #endif // ssrFlag
}
