// Butterfly nebula, created modifying the sources of the Cat's Eye nebula.
// Source: https://www.shadertoy.com/view/XfyyWd
#version 330 core

// Uniforms which are always available
uniform mat4 u_worldTrans;
uniform vec3 u_cameraPos;
uniform float u_bodySize;
uniform vec2 u_cameraNearFar;
uniform float u_cameraK;
uniform float u_time;
uniform float u_simuTime;
// Unit conversion.
uniform float u_kmToU;

#ifdef ssrFlag
#include <shader/lib/ssr.frag.glsl>
#endif // ssrFlag

// INPUT
struct VertexData {
    vec2 texCoords;
    vec3 normal;
    vec3 viewDir;
    vec3 ambientLight;
    float opacity;
    vec4 color;
    vec3 fragPosWorld;
    mat3 tbn;
};
in VertexData v_data;

// OUTPUT
layout(location = 0) out vec4 fragColor;
layout(location = 1) out vec4 layerBuffer;

#include <shader/lib/logdepthbuff.glsl>
#include <$data/catalog-nebulae/shader/lib/nebulae-lib.glsl>

#define DIST_SCALE (u_kmToU * 1.0e6)
#define FACTOR (1.0 / DIST_SCALE * 0.9e-5)

#define ITERATIONS 70.0

float CappedCone(vec3 p, float h, float r1, float r2) {
    vec2 q = vec2(length(p.xz), p.y);
    vec2 k1 = vec2(r2, h);
    vec2 k2 = vec2(r2 - r1, 2.0 * h);
    vec2 ca = vec2(q.x - min(q.x, (q.y < 0.0) ? r1 : r2), abs(q.y) - h);
    vec2 cb = q - k1 + k2 * clamp(dot(k1 - q, k2) / dot2(k2), 0.0, 1.0);
    float s = (cb.x < 0.0 && ca.y < 0.0) ? -1.0 : 1.0;
    return s * sqrt(min(dot2(ca), dot2(cb)));
}

float Xor(float d1, float d2) {
    return max(min(d1, d2), -max(d1, d2));
}
const float nudge = 0.9; // Size of perpendicular vector
float normalizer = 1.0 / sqrt(1.2 + nudge * nudge); // Pythagorean theorem on that perpendicular to maintain scale
float SpiralNoiseC(vec3 p) {
    float n = 0.0; // Noise amount
    float iter = 1.5;
    for (int i = 0; i < 6; i++) {
        // Add sin and cos scaled inverse with the frequency
        n += -abs(sin(p.y * iter) + cos(p.x * iter)) / iter; // Abs for a ridged look
        // Rotate by adding perpendicular and scaling down
        p.xy += vec2(p.y, -p.x) * nudge;
        p.xy *= normalizer;
        // Rotate on other axis
        p.xz += vec2(p.z, -p.x) * nudge;
        p.xz *= normalizer;
        // Increase the frequency
        iter *= 1.633733;
    }
    return n;
}

float NebulaNoise(vec3 p) {
    vec3 p1 = vec3(p.x, p.y - 5.8, p.z);
    float s1 = CappedCone(p1, 5.0, 0.05, 1.4);
    s1 += fbm(p1 * 80.0);
    s1 += SpiralNoiseC(p1 * 0.002);

    vec3 p2 = vec3(p.x, p.y + 6.2, p.z);
    float s2 = CappedCone(p2, -5.0, 0.015, 1.4);
    s2 += fbm(p2 * 80.0);
    s2 += SpiralNoiseC(p2 * 0.001);

    float final = Xor(s2, s1);

    return final;
}

float map(vec3 p) {
    // The Cat's Eye nebula is oriented sideways.
    p = p.yzx;
    vec4 q = get_quat_rotation(vec3(-0.1, 1.0, -0.3), PI / 3.0);
    p = rotate(p, q);
    float NebNoise = abs(NebulaNoise(p * 1.6) * 0.45);
    return NebNoise + 0.086;
}

vec4 raymarch(vec3 camPos, vec3 rayDir, vec3 objPos, vec2 dists, out float s) {
    // Ray direction.
    vec3 rd = rayDir;
    // Ray origin.
    vec3 ro = camPos;

    // ld, td: local, total density
    // w: weighting factor
    float ld = 0.0, td = 0.0, w = 0.0;

    // d: distance function
    float d = 1.0;

    const float h = 0.1;

    vec4 sum = vec4(0.0);

    float min_dist = dists.x, max_dist = dists.y;

    if (min_dist >= 0.0 && max_dist >= 0.0) {
        float t = min_dist;

        // Raymarch loop
        for (int i = 0; i < ITERATIONS; i++) {
            vec3 pos = ro + t * rd;

            // Loop break conditions
            if (td > 0.9 || d < 0.02 * t || t > 30.0 || sum.a > 0.99) break;

            // Evaluate distance function
            float d = map(pos);

            // Change this string to control density
            d = max(d, 0.0);

            // Point light calculations
            vec3 ldst = vec3(0.0) - pos;
            float lDist = max(length(ldst), 0.001);

            // The color of light
            float _T = lDist * 1.6 + 2.6; // <-v endless tweaking
            vec3 lightColor = 0.4 + 0.5 * cos(_T * 0.85 + PI * 0.34 * vec3(0.9, 0.5, 0.7));

            // Star
            sum.rgb += (vec3(0.87, 0.85, 1.0) / (lDist * lDist * 5.0) / 40.0);
            // Bloom
            sum.rgb += (lightColor / exp(lDist * lDist * lDist * 0.2) / 16.0);

            if (d < h) {
                // Compute local density
                ld = h - d;

                // Compute weighting factor
                w = (1.0 - td) * ld;

                // Accumulate density
                td += w + 1.0 / 200.0;

                vec4 col = vec4(0.0, 0.9, 0.8, td);

                // Emission
                sum += sum.a * vec4(sum.rgb, 0.0) * 0.1;

                // Uniform scale density
                col.a *= 0.2;
                // Colour by alpha
                col.rgb *= col.a;
                // Alpha blend in contribution
                sum = sum + col * (1.0 - sum.a);
            }
            td += 1.0 / 90.0;
            // Optimize step size near the camera
            t += max(d * 0.1 * max(min(length(ldst), length(ro)), 1.0), 0.01);
        }
        td *= smoothstep(0.0, 0.8, td);

        // Simple scattering
        sum *= 1. / exp(ld * 0.2) * 0.4;
        sum = clamp(sum, 0.0, 1.0);
        sum.xyz = sum.xyz * sum.xyz * (3.0 - 2.0 * sum.xyz);
    }
    return vec4(desaturate(sum.xyz, 0.1), td);
}

void main() {
    // Directions and positions.
    // Object is alwyas at origin in our reference frame.
    vec3 camPos = u_cameraPos * FACTOR;
    vec3 rayDir = normalize(v_data.fragPosWorld);
    vec3 objPos = vec3(0.0, 0.0, 0.0);
    float radius = u_bodySize;

    vec2 dists = sphereIntersectionDistances(camPos, rayDir, objPos, radius);

    // Ray marching
    float d = 0.0;
    vec4 rmcol = raymarch(camPos, rayDir, objPos, dists, d);
    fragColor = rmcol;

    // Render outline
    // vec2 uv = v_data.texCoords;
    // if (uv.x < 0.005 || uv.x > 0.995 || uv.y < 0.005 || uv.y > 0.995) {
    //     d = length(v_data.fragPosWorld);
    //     fragColor = vec4(1.0, 0.0, 0.0, 1.0);
    // }

    // Logarithmic depth buffer.
    if (d > 0.0) {
        gl_FragDepth = getDepthValue(d, u_cameraNearFar.y, u_cameraK);
    }

    #ifdef ssrFlag
    ssrBuffers();
    #endif // ssrFlag
}
